##
##		This file implements code for
##		mouse-user interaction
##		"Of Mice And Men"
##
##  Copyright (c) 1998 <Irrational Numbers> : <buraldo@maganet.net> 
##
##  This program is free software; you can redistibute it and/or modify
##  it under the terms of the GNU General Public License as published by
##  the Free Software Foundation.
##
##  See the file "license" for information on usage and
##  redistribution of this program, and for a disclaimer of all warranties.
##
##  VER 1.1

namespace eval MUI {

	#
	# constants (they may be changed with Preferences..)
	#
	
set  arc_color(HISELECTED) yellow
set  arc_color(SELECTED) orange
set  arc_color(NORMAL)	black
set node_color(HISELECTED) yellow
set node_color(SELECTED) orange
set node_color(NORMAL)	gray75
set click(gridX) 5
set click(gridY) 5


set click(X) 0
set click(Y) 0
set click(abandon) 0
set click(node_picked) 0 ; # group-tag of picked-node ( 0 is nothing )
set click(arc_picked)  0 ; # group-tag of picked-arc  ( 0 is nothing )
set click(node_picked_list) {}
set click(on_object) 0
set click(selecting) 0
set click(rubbering) 0


	  ## -- popup menus on canvas -----------------------------------------

proc arc_initPopup { w } {
	menu $w.arcPopup -tearoff 0
	set m $w.arcPopup
        $m add command -label "Delete Arc" \
		-command "CVS::arc_delete \$w \$MUI::click(arc_picked)"	
        $m add command -label "Flip Arc" \
		-command "CVS::arc_flip \$w \$MUI::click(arc_picked)"	
}

proc node_initPopup { w } {
	menu $w.nodePopup -tearoff 0
	set m $w.nodePopup
	$m add command -label "Add Arc" \
		-command "CVS::new_arc \$w \$MUI::click(node_picked)"
	$m add command -label "Delete Selected Nodes" \
		-command "CVS::node_delete \$w \$MUI::click(node_picked_list) ;
				MUI::de_select \$w"	
}

proc canvas_initPopup { w } {
	menu $w.canvasPopup -tearoff 0
	set m $w.canvasPopup
        $m add command -label "Connection point" \
		-command "CVS::node \$w connectSymb \$MUI::click(X) \$MUI::click(Y)"
        $m add command -label "Condition symbol" \
		-command "CVS::node \$w conditionSymb \$MUI::click(X) \$MUI::click(Y)"
        $m add command -label "HMSC-Ref symbol" \
		-command "CVS::node \$w mscrefSymb \$MUI::click(X) \$MUI::click(Y)"
        $m add command -label "End Symbol" \
		-command "CVS::node \$w endSymb \$MUI::click(X) \$MUI::click(Y)"
}


	  ## -- rubber-line handling -----------------------------------------

proc disable_rubberLine { w } {
	variable click
	bind $w <Motion> {}
	  # dummy setting for breaking tkwait
	set click(rubbering) 0
	set click(node_picked) $click(node_picked)
}

proc enable_rubberLine { w lineID } {
	variable click
	set click(abandon) 0
	set click(rubbering) 1

	set xyxy [$w coords $lineID]

	set x0 [lindex $xyxy 0]
	set y0 [lindex $xyxy 1]
	bind $w <Motion> [concat  $w coords $lineID $x0 $y0 \
						\[ $w canvasx %x $MUI::click(gridX) \]  \
						\[ $w canvasy %y $MUI::click(gridY) \]  \
					]

	### ------------------------------###
	tkwait variable MUI::click(node_picked)
	### ------------------------------###
	disable_rubberLine $w
}



	# create a select box (keep-it off the paper...)
proc create_select_box {w} {
	$w create rectangle -100 -100 -100 -100 -outline red -tags SELECT_BOX
}


proc checkInside { w tag } {
	return [expr [lsearch -exact [$w gettags current] $tag] >=0]
}



	#
	# Bindings collection
	#
proc init { w } {

  arc_initPopup $w
  node_initPopup $w
  canvas_initPopup $w

  focus $w

  $w bind arc <Leave> {
	%W itemconfigure current -fill $MUI::arc_color(NORMAL)
  }
  $w bind arc <Enter> {
	%W itemconfigure current -fill $MUI::arc_color(HISELECTED)
  }

	  ## these bindings are for breaking bind on empty canvas ...
  $w bind all <ButtonPress-1> {
	set MUI::click(on_object) 1
  }
  $w bind all <ButtonRelease-1> {
	set MUI::click(on_object) 0
  }

  bind $w <Escape> { set MUI::click(abandon) 1; MUI::disable_rubberLine %W }


	# -----------------------------------------------------------------

  proc bind_node_Enter { w } {
	$w configure -cursor fleur
	set thisID [$w find withtag current]
	if { ! [hasTag $w $thisID NODE_SELECTED] } {
		$w itemconfigure $thisID -fill $MUI::node_color(HISELECTED)
	}
	$w bind node <Leave>  "MUI::bind_node_Leave $w"
  }
  proc bind_node_Leave { w } {
	$w configure -cursor {}
	set thisID [$w find withtag current]
	if { ! [hasTag $w $thisID NODE_SELECTED] } {
		$w itemconfigure $thisID -fill $MUI::node_color(NORMAL)
	}
  }
  $w bind node <Enter> { MUI::bind_node_Enter %W }

  $w bind node <ButtonPress-1> {
	set w %W ; set x %x ; set y %y
	##set MUI::click(X)  [$w canvasx $x ]
	set MUI::click(X)  [$w canvasx $x $MUI::click(gridX)]
	##set MUI::click(Y)  [$w canvasy $y] 
	set MUI::click(Y)  [$w canvasy $y $MUI::click(gridY)] 
	set thisID [$w find withtag current]
	if { ! [MUI::is_selected $w $thisID] } {
		MUI::de_select $w
		MUI::mark_selected $w $thisID
	}
	set MUI::click(node_picked) [lgrep [$w gettags $thisID] "grp#*"]

	set inner_arcs {}
	set MUI::crossing_arcs {}
	foreach g $MUI::click(node_picked_list) {
		foreach a [set CVS::$g.arc(out)] {
			if {[lsearch -exact $MUI::click(node_picked_list) [set CVS::$a.node(in)]] >= 0} {
				lappend inner_arcs $a
			} else {
				lappend MUI::crossing_arcs $a
			}
		}
		foreach a [set CVS::$g.arc(in)] {
			if {[lsearch -exact $MUI::click(node_picked_list) [set CVS::$a.node(out)]] >= 0} {
				lappend inner_arcs $a
			} else {
				lappend MUI::crossing_arcs $a
			}
		}
	}
		## remove duplicates from "inner_arcs" list
		## NOTE: crossing_arcs cannot have duplicates
	set inner_arcs [lunique $inner_arcs]
	foreach a $inner_arcs {
		$w addtag SELECTED withtag $a
	}
	
	set xyxy [$w bbox NODE_SELECTED]
	set MUI::xmin  [expr $CVS::XMIN + ($MUI::click(X) - [lindex $xyxy 0])] 
	set MUI::xmax  [expr $CVS::XMAX - ([lindex $xyxy 2] - $MUI::click(X))]
	set MUI::ymin  [expr $CVS::YMIN + ($MUI::click(Y) - [lindex $xyxy 1])] 
	set MUI::ymax  [expr $CVS::YMAX - ([lindex $xyxy 3] - $MUI::click(Y))]

	$w bind node <Enter> {}
	$w bind node <Leave> {}
	$w delete RESIZE_SPOT
	  # node picked! this set stops the rubberLine
	set MUI::click(node_picked) $MUI::click(node_picked)
  }

  $w bind node <B1-Motion> {
	set w %W ; set x %x ; set y %y
	set x [$w canvasx $x $MUI::click(gridX)]
	set y [$w canvasy $y $MUI::click(gridY)]

	if {$x<$MUI::xmin} {set x $MUI::xmin}
	if {$x>$MUI::xmax} {set x $MUI::xmax}
	if {$y<$MUI::ymin} {set y $MUI::ymin}
	if {$y>$MUI::ymax} {set y $MUI::ymax}
	set dx [expr $x - $MUI::click(X)]
	set dy [expr $y - $MUI::click(Y)]
	  ## move !
	$w move SELECTED $dx $dy
	foreach a $MUI::crossing_arcs {
		CVS::arc_redraw $w $a
	}
	set MUI::click(X) $x
	set MUI::click(Y) $y
  }

  $w bind node <ButtonRelease-1> {
	set w %W
	MUI::bind_node_Leave $w
	$w bind node <Enter>  "MUI::bind_node_Enter $w" 
	if { [MUI::checkInside $w node] } { MUI::bind_node_Enter $w }
  }


	# -------------------------------------------------------------------


  proc de_select {w} {
	set MUI::click(selecting) 0
	set MUI::click(node_picked_list) {}
	set MUI::click(node_picked) 0
	$w delete RESIZE_SPOT
	$w coords SELECT_BOX -100 -100 -100 -100
	$w itemconfig NODE_SELECTED -fill $MUI::node_color(NORMAL)
	$w dtag NODE_SELECTED
	$w dtag SELECTED
  }

  proc is_selected {w id} {
	return [hasTag $w $id NODE_SELECTED]
  }

  proc mark_selected {w id} {
	set nodeGrp  [lgrep [$w gettags $id] "grp#*"]
	lappend MUI::click(node_picked_list) $nodeGrp
	$w addtag SELECTED withtag $nodeGrp
	$w itemconfig $id -fill $MUI::node_color(SELECTED)
	$w addtag NODE_SELECTED withtag $id
  }

  bind .c <ButtonPress-1> {
	set w %W ; set x %x ; set y %y
	if { ! $MUI::click(on_object) && ! $MUI::click(rubbering) } {
		MUI::de_select $w
		set MUI::click(selecting) 1

		set MUI::selbox(X0) [$w canvasx $x]
		set MUI::selbox(Y0) [$w canvasy $y]
	}
  }

  bind .c <B1-Motion> {
	set w %W ; set x %x ; set y %y
	if { $MUI::click(selecting) } {
		set MUI::selbox(X1) [$w canvasx $x]
		set MUI::selbox(Y1) [$w canvasy $y]
		$w coords SELECT_BOX  $MUI::selbox(X0) $MUI::selbox(Y0) \
			$MUI::selbox(X1) $MUI::selbox(Y1)
	}
  }

  bind .c <ButtonRelease-1> {
	set w %W ; set x %x ; set y %y
	if { $MUI::click(selecting) } {
		set MUI::selbox(X1) [$w canvasx $x]
		set MUI::selbox(Y1) [$w canvasy $y]
		MUI::de_select $w

		set Lin [$w find enclosed $MUI::selbox(X0) $MUI::selbox(Y0) \
			$MUI::selbox(X1) $MUI::selbox(Y1)]
		 # mark  nodes only
		foreach id $Lin {
			if { [hasTag $w $id node] } {
				MUI::mark_selected $w $id
			}
		}
	}
  }

	# -------------------------------------------------------------------

	## the resizeable node is always the picked-node!

  $w bind resizeableNode <ButtonPress-1> {
	set w %W
	set xyxy [$w bbox withtag current]
	set x2 [lindex $xyxy 2]
	set y2 [lindex $xyxy 3]
	$w delete RESIZE_SPOT
	 ## hotspot should be within bbox (y-coords), or the bbox will be wrong!
	$w create oval [expr $x2-6] [expr $y2-6] [expr $x2] [expr $y2] \
		-fill $MUI::node_color(SELECTED) -tags "RESIZE_SPOT SELECTED"
  }

  proc bind_RESIZE_SPOT_Enter { w } {
	$w configure -cursor bottom_right_corner
	$w bind RESIZE_SPOT <Leave> "MUI::bind_node_Leave $w"
  }
  proc bind_RESIZE_SPOT_Leave { w } {
	$w configure -cursor {}
  }
  $w bind RESIZE_SPOT <Enter> { MUI::bind_RESIZE_SPOT_Enter %W }

  $w bind RESIZE_SPOT <ButtonPress-1> {
	set w %W ; set x %x ; set y %y
	set MUI::click(X)  [$w canvasx $x ]
	set MUI::click(Y)  [$w canvasy $y] 
	 set params [CVS::get_node $w $MUI::click(node_picked)] 
	set MUI::click(xC) [lindex $params 1]
	set MUI::click(yC) [lindex $params 2]
	set MUI::click(node_width) [lindex $params 3]
	set MUI::click(node_height) [lindex $params 4]
	$w bind RESIZE_SPOT <Enter> {}
	$w bind RESIZE_SPOT <Leave> {}

	set xyxy [$w bbox $MUI::click(node_picked)]
	set MUI::xmin  $CVS::XMIN 
	set MUI::xmax  $CVS::XMAX
	set MUI::ymin  $CVS::YMIN
	set MUI::ymax  $CVS::YMAX
  }
  $w bind RESIZE_SPOT <B1-Motion> {
	set w %W ; set x %x ; set y %y
	set x [$w canvasx $x $MUI::click(gridX)]
	set y [$w canvasy $y $MUI::click(gridY)]

	if {$x<$MUI::xmin} {set x $MUI::xmin}
	if {$x>$MUI::xmax} {set x $MUI::xmax}
	if {$y<$MUI::ymin} {set y $MUI::ymin}
	if {$y>$MUI::ymax} {set y $MUI::ymax}

	set dx [expr $x - $MUI::click(X)]
	set dy [expr $y - $MUI::click(Y)]
	$w move RESIZE_SPOT $dx $dy

	set MUI::click(node_width) [expr $MUI::click(node_width) +2*$dx]
	set MUI::click(node_height) [expr $MUI::click(node_height) +2*$dy]
	CVS::redraw_sym_node $w $MUI::click(node_picked) \
		[expr abs($MUI::click(node_width))] [expr abs($MUI::click(node_height))]
	CVS::node_move $w $MUI::click(node_picked) 0 0 
	set MUI::click(X) $x
	set MUI::click(Y) $y
  }
  $w bind RESIZE_SPOT <ButtonRelease-1> {
	set w %W 
	MUI::bind_RESIZE_SPOT_Leave $w
	$w bind RESIZE_SPOT <Enter> "MUI::bind_RESIZE_SPOT_Enter $w"
	if { [MUI::checkInside $w RESIZE_SPOT] } { MUI::bind_RESIZE_SPOT_Enter $w }
  }

	# -------------------------------------------------------------------

  bind $w <Button-3> {
	set w %W ; set X %X ; set Y %Y
	set x %x ; set y %y

	set MUI::click(X)  [$w canvasx $x $MUI::click(gridX)]
	set MUI::click(Y)  [$w canvasy $y $MUI::click(gridY)]
	set id [$w find withtag current]
	if { $id == {} } {
		 # .. on empty canvas ...
		tk_popup $w.canvasPopup $X $Y
	} else {
		## .. specific tags bindings ...
		set thisID [$w find withtag current]
		set grpID [lgrep [$w gettags $thisID] "grp#*"]
		if { [hasTag $w $id node] } {
		 	# .. tag node ...
			$w delete RESIZE_SPOT

			if { ! [MUI::is_selected $w $thisID] } {
				MUI::de_select $w
				MUI::mark_selected $w $thisID
			}
			set MUI::click(node_picked) $grpID

			  # entry 0: Add arc (not on endSymb)
			if { [CVS::get_kind $w $MUI::click(node_picked)] == "endSymb" } {
				$w.nodePopup entryconfigure 0 -state disabled
			} else {
				$w.nodePopup entryconfigure 0 -state normal
			}
			  # entry 1: Delete (not on startSymb)
			if { [CVS::get_kind $w $MUI::click(node_picked)] == "startSymb" } {
				$w.nodePopup entryconfigure 1 -state disabled
			} else {
				$w.nodePopup entryconfigure 1 -state normal
			}
			tk_popup $w.nodePopup $X $Y
		} elseif { [hasTag $w $id arc] } {
		 	# .. tag arc ...
			set MUI::click(arc_picked) $grpID
			  # entry 1: flip arc (not on not-curve)
			if { ! [CVS::arc_is_curve $MUI::click(arc_picked)] } {
				$w.arcPopup entryconfigure 1 -state disabled
			} else {
				$w.arcPopup entryconfigure 1 -state normal
			}
			tk_popup $w.arcPopup $X $Y
		}
	}
  }


}

} ; ## = end of namespace =====================================================

